import os
from advertools.sitemaps import sitemap_to_df
import pandas as pd
import pytest


def full_path(file):
    tests_dir = 'tests/sitemap_testing/'
    return 'file://' + os.path.abspath(tests_dir + file)

regular_sitemap_url = full_path('regular_sitemap.xml')
zipped_sitemap_url = full_path('zipped_sitemap.xml.gz')
sitemap_index_url = full_path('sitemap_index.xml')
error_sitemap_url = full_path('error_sitemap.xml')
image_sitemap_url = full_path('image_sitemap.xml')
video_sitemap_url = full_path('video_sitemap.xml')
news_sitemap_url = full_path('news_sitemap.xml')
robotstxt_url = full_path('robots.txt')


def test_regular_sitemap():
    result = sitemap_to_df(regular_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 5


def test_gz_sitemap():
    result = sitemap_to_df(zipped_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 5


def test_sitemap_index():
    result = sitemap_to_df(sitemap_index_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all([col in result
               for col in ['loc', 'sitemap_downloaded', 'sitemap']])


def test_error_sitemap():
    with pytest.raises(Exception):
        result = sitemap_to_df(error_sitemap_url)


def test_image_sitemap():
    result = sitemap_to_df(image_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert 'image' in result


def test_video_sitemap():
    result = sitemap_to_df(video_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert 'video_content_loc' in result


def test_news_sitemap():
    result = sitemap_to_df(news_sitemap_url)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert 'news' in result


def test_get_sitemaps_from_robotstxt():
    result = sitemap_to_df(robotstxt_url)


import os

import pytest

import advertools.youtube as yt

youtube_key = os.environ.get('GOOG_CSE_KEY')


def test_activities_list():
    result = yt.activities_list(key=youtube_key, part='snippet',
                                channelId='UCv002AUCZaPNwiADqwchijg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_captions_list():
    result = yt.captions_list(key=youtube_key, part='snippet',
                              videoId='kJQP7kiw5Fk')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_channel_sections_list():
    result = yt.channel_sections_list(key=youtube_key, part='snippet',
                                      channelId='UCv002AUCZaPNwiADqwchijg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_channels_list():
    result = yt.channels_list(key=youtube_key, part='snippet,statistics',
                              forUsername='youtube')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_comment_threads_list():
    result = yt.comment_threads_list(key=youtube_key,
                                     part='id,replies,snippet',
                                     videoId='PscrVidwxMg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_comments_list():
    result = yt.comments_list(key=youtube_key, part='snippet',
                              id='UgxKMCc9z4iE7LNW2Hh4AaABAg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_guide_categories_list():
    result = yt.guide_categories_list(key=youtube_key, part='snippet',
                                      regionCode='tr', hl='tr')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_i18n_languages_list():
    result = yt.i18n_languages_list(key=youtube_key, part='snippet')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_i18n_regions_list():
    result = yt.i18n_regions_list(key=youtube_key, part='snippet')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_playlist_items_list():
    result = yt.playlist_items_list(key=youtube_key, part='snippet',
                                    playlistId='PLW0Gy9pTgVntoeYT50HfV144rzuJcrmMg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_playlists_list():
    result = yt.playlists_list(key=youtube_key, part='snippet',
                               channelId='UCv002AUCZaPNwiADqwchijg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_search():
    result = yt.search(key=youtube_key, part='snippet', q='test bitcoin')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_subscriptions_list():
    result = yt.subscriptions_list(key=youtube_key, part='snippet',
                                   channelId='UCv002AUCZaPNwiADqwchijg')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_video_categories_list():
    result = yt.video_categories_list(key=youtube_key, part='snippet',
                                      regionCode='de')
    assert {'queryTime', 'param_part'}.issubset(result.columns)


def test_videos_list():
    result = yt.videos_list(key=youtube_key, part='snippet',
                            chart='mostPopular', regionCode='GB', maxResults=9)
    assert {'queryTime', 'param_part'}.issubset(result.columns)

# Test raising errors:


def test_activities_list_raises():
    with pytest.raises(ValueError):
        yt.activities_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.activities_list(key=youtube_key, part='snippet')


def test_captions_list_raises():
    with pytest.raises(ValueError):
        yt.captions_list(key=youtube_key, part='wrong_part', videoId='random')


def test_channel_sections_list_raises():
    with pytest.raises(ValueError):
        yt.channel_sections_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.channel_sections_list(key=youtube_key, part='snippet')


def test_channels_list_raises():
    with pytest.raises(ValueError):
        yt.channels_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.channels_list(key=youtube_key, part='snippet')


def test_comment_threads_list_raises():
    with pytest.raises(ValueError):
        yt.comment_threads_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.comment_threads_list(key=youtube_key, part='snippet')


def test_comments_list_raises():
    with pytest.raises(ValueError):
        yt.comments_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.comments_list(key=youtube_key, part='snippet')


def test_guide_categories_list_raises():
    with pytest.raises(ValueError):
        yt.guide_categories_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.guide_categories_list(key=youtube_key, part='snippet')


def test_playlist_items_list_raises():
    with pytest.raises(ValueError):
        yt.playlist_items_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.playlist_items_list(key=youtube_key, part='snippet')


def test_playlists_list_raises():
    with pytest.raises(ValueError):
        yt.playlists_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.playlists_list(key=youtube_key, part='snippet')


def test_subscriptions_list_raises():
    with pytest.raises(ValueError):
        yt.subscriptions_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.subscriptions_list(key=youtube_key, part='snippet')


def test_video_categories_list_raises():
    with pytest.raises(ValueError):
        yt.video_categories_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.video_categories_list(key=youtube_key, part='snippet')


def test_videos_list_raises():
    with pytest.raises(ValueError):
        yt.videos_list(key=youtube_key, part='wrong_part')

    with pytest.raises(ValueError):
        yt.videos_list(key=youtube_key, part='snippet')


def test_with_zero_results_required():
    result = yt.search(key=youtube_key, part='snippet', q='testing bitcoin',
                       maxResults=0)


def test_empty_list_raises_error():
    with pytest.raises(ValueError):
        yt.search(key=youtube_key, part='snippet', q='testing bitcoin',
                  regionCode=[])


def test_errors_returned_as_df():
    result = yt.search(key='wrong key', part='snippet', q='testing bitcoin')



"""
.. _ad_create:

Create Ads on a Large Scale
===========================

When creating large-scale campaigns, you also need to create ads on a large
scale. For products in a similar category you typically want to use the same
ads, but only replace the product name, "Get the latest <product> now", and
replace `product` as many times as you have ads.

>>> products = ['Dubai', 'Tokyo', 'Singapore']
>>> adv.ad_create(template='5-star Hotels in {}',
...               replacements=products,
...               max_len=30,
...               fallback='Great Cities')
['5-star Hotels In Dubai',
 '5-star Hotels In Tokyo',
 '5-star Hotels In Singapore']

An important thing to to watch out for, is long product names. Since text ads
have limits on each slot, you need to make sure you don't exceed that limit.
For this you need to provide a `fallback` text in case the product name is
longer than `max_len`.

>>> products = ['Dubai', 'Tokyo', 'Singapore', 'Llanfairpwllgwyngyll']
>>> adv.ad_create(template='5-star Hotels in {}',
...               replacements=products,
...               max_len=30,
...               fallback='Great Cities')
['5-star Hotels In Dubai',
 '5-star Hotels In Tokyo',
 '5-star Hotels In Singapore',
 '5-star Hotels In Great Cities']

"""
import string


def ad_create(template, replacements, fallback, max_len=30, capitalize=True):
    """Insert each of the replacement strings in its place within template.

    :param str template: a string format template, using braces e.g. "Get the
                         latest {} today."
    :param list replacements: replacement strings to be inserted in
                              :attr:`template`
    :param str fallback: the string to insert in :attr:`template` in case
                         :attr:`replacement` is longer than :attr:`max_len`
    :param int max_len: the maximum allowed length of the full string
    :param bool capitalize: whether or not to capitalize words in the result
    :returns formatted: list of strings

    >>> ad_create("Let\'s count {}", ['one', 'two', 'three'], 'one', 20)
    ["Let's Count One", "Let's Count Two", "Let's Count Three"]

    >>> ad_create(template='My favorite car is {}',
    ...           replacements=['Toyota', 'BMW', 'Mercedes', 'Lamborghini'],
    ...           fallback='great',
    ...           max_len=28)
    ['My Favorite Car Is Toyota', 'My Favorite Car Is Bmw',
    'My Favorite Car Is Mercedes', 'My Favorite Car Is Great']

    >>> ad_create('KeEP cApITalization {}', ['As IS'],
    ...           fallback='fallback', max_len=50, capitalize=False)
    ['KeEP cApITalization As IS']

    >>> ad_create('This is very long and will produce and error',
    ...           replacements=['something', 'long'], fallback='Very long',
    ...           max_len=20)
    Traceback (most recent call last):
    File "<input>", line 1, in <module>
    File "<input>", line 26, in ad_create
    ValueError: template + fallback should be <= 20 chars
    """
    if len(template.format(fallback)) > max_len:
        raise ValueError('template + fallback should be <= '
                         + str(max_len) + ' chars')
    final_ad = []
    for rep in replacements:
        if len(template.format(rep)) <= max_len:
            final_ad.append(template.format(rep))
        else:
            final_ad.append(template.format(fallback))

    return [string.capwords(s) for s in final_ad] if capitalize else final_ad



"""
.. _urlytics:

Split, Parse, and Analyze URLs
==============================

Extracting information from URLs can be a little tedious, yet very important.
Using the standard for URLs we can extract a lot of information in a fairly
structured manner.

There are many situations in which you have many URLs that you want to better
understand:

* **Analytics reports**: Whichever analytics sytem you use, whether Google
  Analytics, search console, or any other reporting tool that reports on URLs,
  your reports can be enhanced by splitting URLs, and in effect becoming four
  or five data points as opposed to one.
* :ref:`Crawl datasets <crawl>`: The result of any crawl you run typically
  contains the URLs, which can benefit from the same enhancement.
* :ref:`SERP datasets <serp>`: Which are basically about URLs.
* :ref:`Extracted URLs <extract>`: Extracting URLs from social media posts is
  one thing you might want to do to better understand those posts, and further
  splitting URLs can also help.
* :ref:`XML sitemaps <sitemaps>`: Right after downloading a sitemap(s)
  splitting it further can help in giving a better perspective on the dataset.

The main function here is :func:`url_to_df`, which as the name suggests,
converts URLs to DataFrames.

>>> urls ['https://net.location.com/path_1/path_2?price=10&color=blue#frag_1',
...       'https://net.location.com/path_1/path_2?price=15&color=red#frag_2']
>>> url_to_df(urls)
	                                                              url  scheme	          netloc	          path	                 query	fragment     dir_1	 dir_2	query_price	query_color
0	https://net.location.com/path_1/path_2?price=10&color=blue#frag_1	https	net.location.com	/path_1/path_2	   price=10&color=blue	  frag_1	path_1	path_2	         10	       blue
1	 https://net.location.com/path_1/path_2?price=15&color=red#frag_2	https	net.location.com	/path_1/path_2	    price=15&color=red	  frag_2	path_1	path_2	         15	        red


* **url**: The original URLs are listed as a reference. They are decoded for
  easier reading, and you can set ``decode=False`` if you want to retain the
  original encoding.
* **scheme**: Self-explanatory. Note that you can also provide relative URLs
  `/category/sub-category?one=1&two=2` in which case the `url`, `scheme` and
  `netloc` columns would be empty. You can mix relative and absolute URLs as
  well.
* **netloc**: The network location is the sub-domain (optional) together with
  the domain and top-level domain and/or the country domain.
* **path**: The slug of the URL, excluding the query parameters and fragments
  if any. The path is also split in to directories `dir_1, dir_2, dir_3...` to
  make it easier to categorize and analyze the URLs.
* **query**: If query parameters are available they are given in this column,
  but more importantly they are parsed and included in separate columns, where
  each parameter has its own column (with the keys being the names). As in the
  example above, the query `price=10&color=blue` becomes two columns, one for
  price and the other for color. If any other URLs in the dataset contain the
  same parameters, their values will be populated in the same column, and `NA`
  otherwise.
* **fragment**: The final part of the URL after the hash mark `#`, linking to a
  part in the page.
* **query_***: The query parameter names are prepended with `query_` to make
  it easy to filter them out, and to avoid any name collissions with other
  columns, if some URL contains a query parameter called "url" for example.
  In the unlikely event of having a repeated parameter in the same URL, then
  their values would be delimited by two "@" signs `one@@two@@three`. It's
  unusual, but it happens.
* **hostname and port**: If available a column for ports will be shown, and if
  the hostname is different from `netloc` it would also have its own column.

Query Parameters
----------------
The great thing about parameters is that the names are descriptive (mostly!)
and once given a certain column you can easily understand what data they
contain. Once this is done, you can sort the products by price, filter by
destination, get the red and blue items, and so on.

The URL Path (Directories):
---------------------------
Here things are not as straightforward, and there is no way to know what the
first or second directory is supposed to indicate. In general, I can think of
three main situations that you can encounter while analyzing directories.

* **Consistent URLs**: This is the simplest case, where all URLs follow the
  same structure. `/en/product1` clearly shows that the first directory
  indicates the language of the page. So it can also make sense to rename those
  columns once you have discovered their meaning.

* **Inconsistent URLs**: This is similar to the previous situation. All URLs
  follow the same pattern with a few exceptions. Take the following URLs for
  example:

    * /topic1/title-of-article-1
    * /es/topic1/title-of-article-2
    * /es/topic2/title-of-article-3
    * /topic2/title-of-artilce-4

  You can see that they follow the pattern `/language/topic/article-title`,
  except for English, which is not explicitly mentioned, but its articles can
  be identified by having two instead of three directories, as we have for
  "/es/". If URLs are split in this case, yout will end up with `dir_1` having
  "topic1", "es", "es", and "topic2", which distorts the data. Actually you
  want to have "en", "es", "es", "en". In such cases, after making sure you
  have the right rules and patterns, you might create special columns or
  replace/insert values to make them consistent, and get them to a state
  similar to the first example.

* **URLs of different types**: In many cases you will find that sites having
  different types of pages with completely different roles on the site.

    * /blog/post-1-title.html
    * /community/help/topic_1
    * /community/help/topic_2

  Here, once you split the directories, you will see that they don't align
  properly (because of different lengths), and they can't be compared easily. A
  good approach is to split your dataset into one for blog posts and another
  for community content for example.

The ideal case for the `path` part of the URL is to be split into directories
of equal length across the dataset, having the right data in the right columns
and `NA` otherwise. Or, splitting the dataset and analyzing separately.
"""
from urllib.parse import urlsplit, parse_qs, unquote

import pandas as pd


def url_to_df(urls, decode=True):
    """Split the given URLs into their components to a DataFrame.

    Each column will have its own component, and query parameters and
    directories will also be parsed and given special columns each.

    :param url urls: A list of URLs to split into components
    :param bool decode: Whether or not to decode the given URLs
    :return DataFrame split: A DataFrame with a column for each component
    """
    if isinstance(urls, str):
        urls = [urls]
    decode = unquote if decode else lambda x: x
    split_list = []
    for url in urls:
        split = urlsplit(decode(url))
        port = split.port
        hostname = split.hostname if split.hostname != split.netloc else None
        split = split._asdict()
        if hostname:
            split['hostname'] = hostname
        if port:
            split['port'] = port
        parsed_query = parse_qs(split['query'])
        parsed_query = {'query_' + key: '@@'.join(val)
                        for key, val in parsed_query.items()}
        split.update(**parsed_query)
        dirs = split['path'].strip('/').split('/')
        if dirs[0]:
            dir_cols = {'dir_{}'.format(n): d for n, d in enumerate(dirs, 1)}
            split.update(**dir_cols)
        split_list.append(split)
    df = pd.DataFrame(split_list)

    query_df = df.filter(regex='query_')
    if not query_df.empty:
        df = df.drop(query_df.columns, axis=1)
    dirs_df = df.filter(regex='dir_')
    if not dirs_df.empty:
        df = df.drop(dirs_df.columns, axis=1)
    df = pd.concat([df, dirs_df, query_df], axis=1)
    df.insert(0, 'url', [decode(url) for url in urls])
    return df



"""
.. _youtube:

YouTube Data API
================
"""

from ._yt_helpers import _combine_requests

__all__ = [
    'activities_list', 'captions_list', 'channel_sections_list',
    'channels_list', 'comment_threads_list', 'comments_list',
    'guide_categories_list', 'i18n_languages_list', 'i18n_regions_list',
    'playlist_items_list', 'playlists_list', 'search', 'subscriptions_list',
    'video_categories_list', 'videos_list',
]

def activities_list(key, part, channelId=None, home=None, mine=None,
                    maxResults=None, pageToken=None, publishedAfter=None,
                    publishedBefore=None, regionCode=None):
    """Returns a list of channel activity events that match the request
    criteria. For example, you can retrieve events associated with a particular
    channel or with the user's own channel.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more activity resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in an activity resource, the snippet property contains other
        properties that identify the type of activity, a display title for the
        activity, and so forth. If you set part=snippet, the API response will
        also contain all of those nested properties.The following list contains
        the part names that you can include in the parameter value and the
        quota cost for each part: contentDetails: 2 id: 0 snippet: 2

    *Filters (specify exactly one of the following parameters):*

    :param channelId: string  The channelId parameter specifies a unique
        YouTube channel ID. The API will then return a list of that channel's
        activities.
    :param home: boolean  Note: This parameter has been deprecated.For requests
        that set this parameter, the API response contains items similar to
        those that a logged-out user would see on the YouTube home page. Note
        that this parameter can only be used in a properly authorized request.
    :param mine: boolean  This parameter can only be used in a properly
        authorized request. Set this parameter's value to true to retrieve a
        feed of the authenticated user's activities.

    *Optional parameters:*

    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    :param publishedAfter: datetime  The publishedAfter parameter specifies the
        earliest date and time that an activity could have occurred for that
        activity to be included in the API response. If the parameter value
        specifies a day, but not a time, then any activities that occurred that
        day will be included in the result set. The value is specified in ISO
        8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
    :param publishedBefore: datetime  The publishedBefore parameter specifies
        the date and time before which an activity must have occurred for that
        activity to be included in the API response. If the parameter value
        specifies a day, but not a time, then any activities that occurred that
        day will be excluded from the result set. The value is specified in ISO
        8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
    :param regionCode: string  The regionCode parameter instructs the API to
        return results for the specified country. The parameter value is an ISO
        3166-1 alpha-2 country code. YouTube uses this value when the
        authorized user's previous activity on YouTube does not provide enough
        information to generate the activity feed.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'snippet'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [channelId, home, mine]]) != 1:
        raise ValueError("make sure you specify exactly one of ['channelId', 'home', 'mine']")

    base_url = 'https://www.googleapis.com/youtube/v3/activities'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def videos_list(key, part, chart=None, id=None, myRating=None, hl=None,
                maxHeight=None, maxResults=None, maxWidth=None,
                onBehalfOfContentOwner=None, pageToken=None, regionCode=None,
                videoCategoryId=None):
    """Returns a list of videos that match the API request parameters.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more video resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a video resource, the snippet property contains the
        channelId, title, description, tags, and categoryId properties. As
        such, if you set part=snippet, the API response will contain all of
        those properties.The following list contains the part names that you
        can include in the parameter value and the quota cost for each part:
        contentDetails: 2 fileDetails: 1 id: 0 liveStreamingDetails: 2
        localizations: 2 player: 0 processingDetails: 1 recordingDetails: 2
        snippet: 2 statistics: 2 status: 2 suggestions: 1 topicDetails: 2

    *Filters (specify exactly one of the following parameters):*

    :param chart: string  The chart parameter identifies the chart that you
        want to retrieve.Acceptable values are: mostPopular – Return the most
        popular videos for the specified content region and video category.
    :param id: string  The id parameter specifies a comma-separated list of the
        YouTube video ID(s) for the resource(s) that are being retrieved. In a
        video resource, the id property specifies the video's ID.
    :param myRating: string  This parameter can only be used in a properly
        authorized request. Set this parameter's value to like or dislike to
        instruct the API to only return videos liked or disliked by the
        authenticated user.Acceptable values are: dislike – Returns only videos
        disliked by the authenticated user. like – Returns only video liked by
        the authenticated user.

    *Optional parameters:*

    :param hl: string  The hl parameter instructs the API to retrieve localized
        resource metadata for a specific application language that the YouTube
        website supports. The parameter value must be a language code included
        in the list returned by the i18nLanguages.list method.If localized
        resource details are available in that language, the resource's
        snippet.localized object will contain the localized values. However, if
        localized details are not available, the snippet.localized object will
        contain resource details in the resource's default language.
    :param maxHeight: unsigned integer  The maxHeight parameter specifies the
        maximum height of the embedded player returned in the player.embedHtml
        property. You can use this parameter to specify that instead of the
        default dimensions, the embed code should use a height appropriate for
        your application layout. If the maxWidth parameter is also provided,
        the player may be shorter than the maxHeight in order to not violate
        the maximum width. Acceptable values are 72 to 8192, inclusive.
    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param maxWidth: unsigned integer  The maxWidth parameter specifies the
        maximum width of the embedded player returned in the player.embedHtml
        property. You can use this parameter to specify that instead of the
        default dimensions, the embed code should use a width appropriate for
        your application layout.If the maxHeight parameter is also provided,
        the player may be narrower than maxWidth in order to not violate the
        maximum height. Acceptable values are 72 to 8192, inclusive.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.Note: This parameter is supported for use in
        conjunction with the myRating parameter, but it is not supported for
        use in conjunction with the id parameter.
    :param regionCode: string  The regionCode parameter instructs the API to
        select a video chart available in the specified region. This parameter
        can only be used in conjunction with the chart parameter. The parameter
        value is an ISO 3166-1 alpha-2 country code.
    :param videoCategoryId: string  The videoCategoryId parameter identifies
        the video category for which the chart should be retrieved. This
        parameter can only be used in conjunction with the chart parameter. By
        default, charts are not restricted to a particular category. The
        default value is 0.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'processingDetails', 'fileDetails', 'snippet', 'localizations', 'suggestions', 'statistics', 'liveStreamingDetails', 'player', 'status', 'recordingDetails', 'topicDetails'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [chart, id, myRating]]) != 1:
        raise ValueError("make sure you specify exactly one of ['chart', 'id', 'myRating']")

    base_url = 'https://www.googleapis.com/youtube/v3/videos'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def video_categories_list(key, part, id=None, regionCode=None, hl=None):
    """Returns a list of categories that can be associated with YouTube videos.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies the videoCategory
        resource properties that the API response will include. Set the
        parameter value to snippet. The snippet part has a quota cost of 2
        units.

    *Filters (specify exactly one of the following parameters):*

    :param id: string  The id parameter specifies a comma-separated list of
        video category IDs for the resources that you are retrieving.
    :param regionCode: string  The regionCode parameter instructs the API to
        return the list of video categories available in the specified country.
        The parameter value is an ISO 3166-1 alpha-2 country code.

    *Optional parameters:*

    :param hl: string  The hl parameter specifies the language that should be
        used for text values in the API response. The default value is en_US.
    """
    args = locals()
    if sum([bool(p) for p in [id, regionCode]]) != 1:
        raise ValueError("make sure you specify exactly one of ['id', 'regionCode']")

    base_url = 'https://www.googleapis.com/youtube/v3/videoCategories'
    return _combine_requests(args, base_url, count=None, max_allowed=None)


def search(key, part, forContentOwner=None, forDeveloper=None,
           forMine=None, relatedToVideoId=None, channelId=None,
           channelType=None, eventType=None, location=None,
           locationRadius=None, maxResults=None,
           onBehalfOfContentOwner=None, order=None, pageToken=None,
           publishedAfter=None, publishedBefore=None, q=None,
           regionCode=None, relevanceLanguage=None, safeSearch=None,
           topicId=None, type=None, videoCaption=None,
           videoCategoryId=None, videoDefinition=None,
           videoDimension=None, videoDuration=None, videoEmbeddable=None,
           videoLicense=None, videoSyndicated=None, videoType=None):
    """Returns a collection of search results that match the query parameters
    specified in the API request. By default, a search result set identifies
    matching  video ,  channel , and  playlist  resources, but you can also
    configure queries to only retrieve a specific type of resource.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more search resource properties that the API response will
        include. Set the parameter value to snippet.

    *Filters (specify 0 or 1 of the following parameters):*

    :param forContentOwner: boolean  This parameter can only be used in a
        properly authorized request, and it is intended exclusively for YouTube
        content partners.The forContentOwner parameter restricts the search to
        only retrieve videos owned by the content owner identified by the
        onBehalfOfContentOwner parameter. If forContentOwner is set to true,
        the request must also meet these requirements:The
        onBehalfOfContentOwner parameter is required.The user authorizing the
        request must be using an account linked to the specified content
        owner.The type parameter value must be set to video.None of the
        following other parameters can be set: videoDefinition, videoDimension,
        videoDuration, videoLicense, videoEmbeddable, videoSyndicated,
        videoType.
    :param forDeveloper: boolean  This parameter can only be used in a properly
        authorized request. The forDeveloper parameter restricts the search to
        only retrieve videos uploaded via the developer's application or
        website. The API server uses the request's authorization credentials to
        identify the developer. The forDeveloper parameter can be used in
        conjunction with optional search parameters like the q parameter.For
        this feature, each uploaded video is automatically tagged with the
        project number that is associated with the developer's application in
        the Google Developers Console.When a search request subsequently sets
        the forDeveloper parameter to true, the API server uses the request's
        authorization credentials to identify the developer. Therefore, a
        developer can restrict results to videos uploaded through the
        developer's own app or website but not to videos uploaded through other
        apps or sites.
    :param forMine: boolean  This parameter can only be used in a properly
        authorized request. The forMine parameter restricts the search to only
        retrieve videos owned by the authenticated user. If you set this
        parameter to true, then the type parameter's value must also be set to
        video. In addition, none of the following other parameters can be set
        in the same request: videoDefinition, videoDimension, videoDuration,
        videoLicense, videoEmbeddable, videoSyndicated, videoType.
    :param relatedToVideoId: string  The relatedToVideoId parameter retrieves a
        list of videos that are related to the video that the parameter value
        identifies. The parameter value must be set to a YouTube video ID and,
        if you are using this parameter, the type parameter must be set to
        video.Note that if the relatedToVideoId parameter is set, the only
        other supported parameters are part, maxResults, pageToken, regionCode,
        relevanceLanguage, safeSearch, type (which must be set to video), and
        fields.

    *Optional parameters:*

    :param channelId: string  The channelId parameter indicates that the API
        response should only contain resources created by the channel. Note:
        Search results are constrained to a maximum of 500 videos if your
        request specifies a value for the channelId parameter and sets the type
        parameter value to video, but it does not also set one of the
        forContentOwner, forDeveloper, or forMine filters.
    :param channelType: string  The channelType parameter lets you restrict a
        search to a particular type of channel.Acceptable values are: any –
        Return all channels. show – Only retrieve shows.
    :param eventType: string  The eventType parameter restricts a search to
        broadcast events. If you specify a value for this parameter, you must
        also set the type parameter's value to video.Acceptable values are:
        completed – Only include completed broadcasts. live – Only include
        active broadcasts. upcoming – Only include upcoming broadcasts.
    :param location: string  The location parameter, in conjunction with the
        locationRadius parameter, defines a circular geographic area and also
        restricts a search to videos that specify, in their metadata, a
        geographic location that falls within that area. The parameter value is
        a string that specifies latitude/longitude coordinates e.g.
        (37.42307,-122.08427).The location parameter value identifies the point
        at the center of the area.The locationRadius parameter specifies the
        maximum distance that the location associated with a video can be from
        that point for the video to still be included in the search results.The
        API returns an error if your request specifies a value for the location
        parameter but does not also specify a value for the locationRadius
        parameter.
    :param locationRadius: string  The locationRadius parameter, in conjunction
        with the location parameter, defines a circular geographic area.The
        parameter value must be a floating point number followed by a
        measurement unit. Valid measurement units are m, km, ft, and mi. For
        example, valid parameter values include 1500m, 5km, 10000ft, and
        0.75mi. The API does not support locationRadius parameter values larger
        than 1000 kilometers.Note: See the definition of the location parameter
        for more information.
    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param order: string  The order parameter specifies the method that will be
        used to order resources in the API response. The default value is
        relevance.Acceptable values are: date – Resources are sorted in reverse
        chronological order based on the date they were created. rating –
        Resources are sorted from highest to lowest rating. relevance –
        Resources are sorted based on their relevance to the search query. This
        is the default value for this parameter. title – Resources are sorted
        alphabetically by title. videoCount – Channels are sorted in descending
        order of their number of uploaded videos. viewCount – Resources are
        sorted from highest to lowest number of views. For live broadcasts,
        videos are sorted by number of concurrent viewers while the broadcasts
        are ongoing.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    :param publishedAfter: datetime  The publishedAfter parameter indicates
        that the API response should only contain resources created at or after
        the specified time. The value is an RFC 3339 formatted date-time value
        (1970-01-01T00:00:00Z).
    :param publishedBefore: datetime  The publishedBefore parameter indicates
        that the API response should only contain resources created before or
        at the specified time. The value is an RFC 3339 formatted date-time
        value (1970-01-01T00:00:00Z).
    :param q: string  The q parameter specifies the query term to search
        for.Your request can also use the Boolean NOT (-) and OR (|) operators
        to exclude videos or to find videos that are associated with one of
        several search terms. For example, to search for videos matching either
        "boating" or "sailing", set the q parameter value to boating|sailing.
        Similarly, to search for videos matching either "boating" or "sailing"
        but not "fishing", set the q parameter value to boating|sailing
        -fishing. Note that the pipe character must be URL-escaped when it is
        sent in your API request. The URL-escaped value for the pipe character
        is %7C.
    :param regionCode: string  The regionCode parameter instructs the API to
        return search results for videos that can be viewed in the specified
        country. The parameter value is an ISO 3166-1 alpha-2 country code.
    :param relevanceLanguage: string  The relevanceLanguage parameter instructs
        the API to return search results that are most relevant to the
        specified language. The parameter value is typically an ISO 639-1 two-
        letter language code. However, you should use the values zh-Hans for
        simplified Chinese and zh-Hant for traditional Chinese. Please note
        that results in other languages will still be returned if they are
        highly relevant to the search query term.
    :param safeSearch: string  The safeSearch parameter indicates whether the
        search results should include restricted content as well as standard
        content.Acceptable values are: moderate – YouTube will filter some
        content from search results and, at the least, will filter content that
        is restricted in your locale. Based on their content, search results
        could be removed from search results or demoted in search results. This
        is the default parameter value. none – YouTube will not filter the
        search result set. strict – YouTube will try to exclude all restricted
        content from the search result set. Based on their content, search
        results could be removed from search results or demoted in search
        results.
    :param topicId: string  The topicId parameter indicates that the API
        response should only contain resources associated with the specified
        topic. The value identifies a Freebase topic ID.Important: Due to the
        deprecation of Freebase and the Freebase API, the topicId parameter
        started working differently as of February 27, 2017. At that time,
        YouTube started supporting a small set of curated topic IDs, and you
        can only use that smaller set of IDs as values for this parameter.  See
        topic IDs supported as of February 15, 2017  Topics  Music topics
        /m/04rlf  Music (parent topic)  /m/02mscn  Christian music  /m/0ggq0m
        Classical music  /m/01lyv  Country  /m/02lkt  Electronic music
        /m/0glt670  Hip hop music  /m/05rwpb  Independent music  /m/03_d0  Jazz
        /m/028sqc  Music of Asia  /m/0g293  Music of Latin America  /m/064t9
        Pop music  /m/06cqb  Reggae  /m/06j6l  Rhythm and blues  /m/06by7  Rock
        music  /m/0gywn  Soul music  Gaming topics  /m/0bzvm2  Gaming (parent
        topic)  /m/025zzc  Action game  /m/02ntfj  Action-adventure game
        /m/0b1vjn  Casual game  /m/02hygl  Music video game  /m/04q1x3q  Puzzle
        video game  /m/01sjng  Racing video game  /m/0403l3g  Role-playing
        video game  /m/021bp2  Simulation video game  /m/022dc6  Sports game
        /m/03hf_rm  Strategy video game  Sports topics  /m/06ntj  Sports
        (parent topic)  /m/0jm\_  American football  /m/018jz  Baseball
        /m/018w8  Basketball  /m/01cgz  Boxing  /m/09xp\_  Cricket  /m/02vx4
        Football  /m/037hz  Golf  /m/03tmr  Ice hockey  /m/01h7lh  Mixed
        martial arts  /m/0410tth  Motorsport  /m/07bs0  Tennis  /m/07_53
        Volleyball  Entertainment topics  /m/02jjt  Entertainment (parent
        topic)  /m/09kqc  Humor  /m/02vxn  Movies  /m/05qjc  Performing arts
        /m/066wd  Professional wrestling  /m/0f2f9  TV shows  Lifestyle topics
        /m/019_rr  Lifestyle (parent topic)  /m/032tl  Fashion  /m/027x7n
        Fitness  /m/02wbm  Food  /m/03glg  Hobby  /m/068hy  Pets  /m/041xxh
        Physical attractiveness [Beauty]  /m/07c1v  Technology  /m/07bxq
        Tourism  /m/07yv9  Vehicles  Society topics  /m/098wr  Society (parent
        topic)  /m/09s1f  Business  /m/0kt51  Health  /m/01h6rj  Military
        /m/05qt0  Politics  /m/06bvp  Religion  Other topics  /m/01k8wb
        Knowledge
    :param type: string  The type parameter restricts a search query to only
        retrieve a particular type of resource. The value is a comma-separated
        list of resource types. The default value is
        video,channel,playlist.Acceptable values are: channelplaylistvideo
    :param videoCaption: string  The videoCaption parameter indicates whether
        the API should filter video search results based on whether they have
        captions. If you specify a value for this parameter, you must also set
        the type parameter's value to video.Acceptable values are: any – Do not
        filter results based on caption availability. closedCaption – Only
        include videos that have captions. none – Only include videos that do
        not have captions.
    :param videoCategoryId: string  The videoCategoryId parameter filters video
        search results based on their category. If you specify a value for this
        parameter, you must also set the type parameter's value to video.
    :param videoDefinition: string  The videoDefinition parameter lets you
        restrict a search to only include either high definition (HD) or
        standard definition (SD) videos. HD videos are available for playback
        in at least 720p, though higher resolutions, like 1080p, might also be
        available. If you specify a value for this parameter, you must also set
        the type parameter's value to video.Acceptable values are: any – Return
        all videos, regardless of their resolution. high – Only retrieve HD
        videos. standard – Only retrieve videos in standard definition.
    :param videoDimension: string  The videoDimension parameter lets you
        restrict a search to only retrieve 2D or 3D videos. If you specify a
        value for this parameter, you must also set the type parameter's value
        to video.Acceptable values are: 2d – Restrict search results to exclude
        3D videos. 3d – Restrict search results to only include 3D videos. any
        – Include both 3D and non-3D videos in returned results. This is the
        default value.
    :param videoDuration: string  The videoDuration parameter filters video
        search results based on their duration. If you specify a value for this
        parameter, you must also set the type parameter's value to
        video.Acceptable values are: any – Do not filter video search results
        based on their duration. This is the default value. long – Only include
        videos longer than 20 minutes. medium – Only include videos that are
        between four and 20 minutes long (inclusive). short – Only include
        videos that are less than four minutes long.
    :param videoEmbeddable: string  The videoEmbeddable parameter lets you to
        restrict a search to only videos that can be embedded into a webpage.
        If you specify a value for this parameter, you must also set the type
        parameter's value to video.Acceptable values are: any – Return all
        videos, embeddable or not. true – Only retrieve embeddable videos.
    :param videoLicense: string  The videoLicense parameter filters search
        results to only include videos with a particular license. YouTube lets
        video uploaders choose to attach either the Creative Commons license or
        the standard YouTube license to each of their videos. If you specify a
        value for this parameter, you must also set the type parameter's value
        to video.Acceptable values are: any – Return all videos, regardless of
        which license they have, that match the query parameters.
        creativeCommon – Only return videos that have a Creative Commons
        license. Users can reuse videos with this license in other videos that
        they create. Learn more. youtube – Only return videos that have the
        standard YouTube license.
    :param videoSyndicated: string  The videoSyndicated parameter lets you to
        restrict a search to only videos that can be played outside
        youtube.com. If you specify a value for this parameter, you must also
        set the type parameter's value to video.Acceptable values are: any –
        Return all videos, syndicated or not. true – Only retrieve syndicated
        videos.
    :param videoType: string  The videoType parameter lets you restrict a
        search to a particular type of videos. If you specify a value for this
        parameter, you must also set the type parameter's value to
        video.Acceptable values are: any – Return all videos. episode – Only
        retrieve episodes of shows. movie – Only retrieve movies.
    """
    args = locals()

    base_url = 'https://www.googleapis.com/youtube/v3/search'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def subscriptions_list(key, part, channelId=None, id=None, mine=None,
                       myRecentSubscribers=None, mySubscribers=None,
                       forChannelId=None, maxResults=None,
                       onBehalfOfContentOwner=None,
                       onBehalfOfContentOwnerChannel=None, order=None,
                       pageToken=None):
    """Returns subscription resources that match the API request criteria.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more subscription resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a subscription resource, the snippet property contains
        other properties, such as a display title for the subscription. If you
        set part=snippet, the API response will also contain all of those
        nested properties.The following list contains the part names that you
        can include in the parameter value and the quota cost for each part:
        contentDetails: 2 id: 0 snippet: 2 subscriberSnippet: 2

    *Filters (specify exactly one of the following parameters):*

    :param channelId: string  The channelId parameter specifies a YouTube
        channel ID. The API will only return that channel's subscriptions.
    :param id: string  The id parameter specifies a comma-separated list of the
        YouTube subscription ID(s) for the resource(s) that are being
        retrieved. In a subscription resource, the id property specifies the
        YouTube subscription ID.
    :param mine: boolean  This parameter can only be used in a properly
        authorized request. Set this parameter's value to true to retrieve a
        feed of the authenticated user's subscriptions.
    :param myRecentSubscribers: boolean  This parameter can only be used in a
        properly authorized request. Set this parameter's value to true to
        retrieve a feed of the subscribers of the authenticated user in reverse
        chronological order (newest first).Note that this parameter only
        supports retrieval of the most recent 1000 subscribers to the
        authenticated user's channel. To retrieve a complete list of
        subscribers, use the mySubscribers parameter. That parameter, which
        does not return subscribers in a particular order, does not limit the
        number of subscribers that can be retrieved.
    :param mySubscribers: boolean  This parameter can only be used in a
        properly authorized request. Set this parameter's value to true to
        retrieve a feed of the subscribers of the authenticated user in no
        particular order.

    *Optional parameters:*

    :param forChannelId: string  The forChannelId parameter specifies a comma-
        separated list of channel IDs. The API response will then only contain
        subscriptions matching those channels.
    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param onBehalfOfContentOwner: string  Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param onBehalfOfContentOwnerChannel: string  This parameter can only be
        used in a properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The
        onBehalfOfContentOwnerChannel parameter specifies the YouTube channel
        ID of the channel to which a video is being added. This parameter is
        required when a request specifies a value for the
        onBehalfOfContentOwner parameter, and it can only be used in
        conjunction with that parameter. In addition, the request must be
        authorized using a CMS account that is linked to the content owner that
        the onBehalfOfContentOwner parameter specifies. Finally, the channel
        that the onBehalfOfContentOwnerChannel parameter value specifies must
        be linked to the content owner that the onBehalfOfContentOwner
        parameter specifies.This parameter is intended for YouTube content
        partners that own and manage many different YouTube channels. It allows
        content owners to authenticate once and perform actions on behalf of
        the channel specified in the parameter value, without having to provide
        authentication credentials for each separate channel.
    :param order: string  The order parameter specifies the method that will be
        used to sort resources in the API response. The default value is
        SUBSCRIPTION_ORDER_RELEVANCE.Acceptable values are: alphabetical – Sort
        alphabetically. relevance – Sort by relevance. unread – Sort by order
        of activity.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'subscriberSnippet', 'snippet'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [channelId, id, mine, myRecentSubscribers, mySubscribers]]) != 1:
        raise ValueError("make sure you specify exactly one of ['channelId', 'id', 'mine', 'myRecentSubscribers', 'mySubscribers']")

    base_url = 'https://www.googleapis.com/youtube/v3/subscriptions'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def i18n_regions_list(key, part, hl=None):
    """Returns a list of content regions that the YouTube website supports.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies the i18nRegion resource
        properties that the API response will include. Set the parameter value
        to snippet. The snippet part has a quota cost of 1 unit.

    *Optional parameters:*

    :param hl: string  The hl parameter specifies the language that should be
        used for text values in the API response. The default value is en_US.
    """
    args = locals()

    base_url = 'https://www.googleapis.com/youtube/v3/i18nRegions'
    return _combine_requests(args, base_url, count=None, max_allowed=None)


def playlists_list(key, part, channelId=None, id=None, mine=None, hl=None,
                   maxResults=None, onBehalfOfContentOwner=None,
                   onBehalfOfContentOwnerChannel=None, pageToken=None):
    """Returns a collection of playlists that match the API request parameters.
    For example, you can retrieve all playlists that the authenticated user
    owns, or you can retrieve one or more playlists by their unique IDs.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more playlist resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a playlist resource, the snippet property contains
        properties like author, title, description, tags, and timeCreated. As
        such, if you set part=snippet, the API response will contain all of
        those properties.The following list contains the part names that you
        can include in the parameter value and the quota cost for each part:
        contentDetails: 2 id: 0 localizations: 2 player: 0 snippet: 2 status: 2

    *Filters (specify exactly one of the following parameters):*

    :param channelId: string  This value indicates that the API should only
        return the specified channel's playlists.
    :param id: string  The id parameter specifies a comma-separated list of the
        YouTube playlist ID(s) for the resource(s) that are being retrieved. In
        a playlist resource, the id property specifies the playlist's YouTube
        playlist ID.
    :param mine: boolean  This parameter can only be used in a properly
        authorized request. Set this parameter's value to true to instruct the
        API to only return playlists owned by the authenticated user.

    *Optional parameters:*

    :param hl: string  The hl parameter instructs the API to retrieve localized
        resource metadata for a specific application language that the YouTube
        website supports. The parameter value must be a language code included
        in the list returned by the i18nLanguages.list method.If localized
        resource details are available in that language, the resource's
        snippet.localized object will contain the localized values. However, if
        localized details are not available, the snippet.localized object will
        contain resource details in the resource's default language.
    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param onBehalfOfContentOwnerChannel: string  This parameter can only be
        used in a properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The
        onBehalfOfContentOwnerChannel parameter specifies the YouTube channel
        ID of the channel to which a video is being added. This parameter is
        required when a request specifies a value for the
        onBehalfOfContentOwner parameter, and it can only be used in
        conjunction with that parameter. In addition, the request must be
        authorized using a CMS account that is linked to the content owner that
        the onBehalfOfContentOwner parameter specifies. Finally, the channel
        that the onBehalfOfContentOwnerChannel parameter value specifies must
        be linked to the content owner that the onBehalfOfContentOwner
        parameter specifies.This parameter is intended for YouTube content
        partners that own and manage many different YouTube channels. It allows
        content owners to authenticate once and perform actions on behalf of
        the channel specified in the parameter value, without having to provide
        authentication credentials for each separate channel.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'snippet', 'localizations', 'player', 'status'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [channelId, id, mine]]) != 1:
        raise ValueError("make sure you specify exactly one of ['channelId', 'id', 'mine']")

    base_url = 'https://www.googleapis.com/youtube/v3/playlists'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def i18n_languages_list(key, part, hl=None):
    """Returns a list of application languages that the YouTube website
    supports.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies the i18nLanguage resource
        properties that the API response will include. Set the parameter value
        to snippet. The snippet part has a quota cost of 1 unit.

    *Optional parameters:*

    :param hl: string  The hl parameter specifies the language that should be
        used for text values in the API response. The default value is en_US.
    """
    args = locals()

    base_url = 'https://www.googleapis.com/youtube/v3/i18nLanguages'
    return _combine_requests(args, base_url, count=None, max_allowed=None)


def playlist_items_list(key, part, id=None, playlistId=None, maxResults=None,
                        onBehalfOfContentOwner=None, pageToken=None,
                        videoId=None):
    """Returns a collection of playlist items that match the API request
    parameters. You can retrieve all of the playlist items in a specified
    playlist or retrieve one or more playlist items by their unique IDs.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more playlistItem resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a playlistItem resource, the snippet property contains
        numerous fields, including the title, description, position, and
        resourceId properties. As such, if you set part=snippet, the API
        response will contain all of those properties.The following list
        contains the part names that you can include in the parameter value and
        the quota cost for each part: contentDetails: 2 id: 0 snippet: 2
        status: 2

    *Filters (specify exactly one of the following parameters):*

    :param id: string  The id parameter specifies a comma-separated list of one
        or more unique playlist item IDs.
    :param playlistId: string  The playlistId parameter specifies the unique ID
        of the playlist for which you want to retrieve playlist items. Note
        that even though this is an optional parameter, every request to
        retrieve playlist items must specify a value for either the id
        parameter or the playlistId parameter.

    *Optional parameters:*

    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    :param videoId: string  The videoId parameter specifies that the request
        should return only the playlist items that contain the specified video.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'snippet', 'status'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [id, playlistId]]) != 1:
        raise ValueError("make sure you specify exactly one of ['id', 'playlistId']")

    base_url = 'https://www.googleapis.com/youtube/v3/playlistItems'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def guide_categories_list(key, part, id=None, regionCode=None, hl=None):
    """Returns a list of categories that can be associated with YouTube
    channels.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies the guideCategory
        resource properties that the API response will include. Set the
        parameter value to snippet. The snippet part has a quota cost of 2
        units.

    *Filters (specify exactly one of the following parameters):*

    :param id: string  The id parameter specifies a comma-separated list of the
        YouTube channel category ID(s) for the resource(s) that are being
        retrieved. In a guideCategory resource, the id property specifies the
        YouTube channel category ID.
    :param regionCode: string  The regionCode parameter instructs the API to
        return the list of guide categories available in the specified country.
        The parameter value is an ISO 3166-1 alpha-2 country code.

    *Optional parameters:*

    :param hl: string  The hl parameter specifies the language that will be
        used for text values in the API response. The default value is en-US.
    """
    args = locals()
    if sum([bool(p) for p in [id, regionCode]]) != 1:
        raise ValueError("make sure you specify exactly one of ['id', 'regionCode']")

    base_url = 'https://www.googleapis.com/youtube/v3/guideCategories'
    return _combine_requests(args, base_url, count=None, max_allowed=None)


def comment_threads_list(key, part, allThreadsRelatedToChannelId=None,
                         channelId=None, id=None, videoId=None,
                         maxResults=None, moderationStatus=None, order=None,
                         pageToken=None, searchTerms=None, textFormat=None):
    """Returns a list of comment threads that match the API request parameters.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more commentThread resource properties that the API response
        will include.The following list contains the part names that you can
        include in the parameter value and the quota cost for each part: id: 0
        replies: 2 snippet: 2

    *Filters (specify exactly one of the following parameters):*

    :param allThreadsRelatedToChannelId: string  The
        allThreadsRelatedToChannelId parameter instructs the API to return all
        comment threads associated with the specified channel. The response can
        include comments about the channel or about the channel's videos.
    :param channelId: string  The channelId parameter instructs the API to
        return comment threads containing comments about the specified channel.
        (The response will not include comments left on videos that the channel
        uploaded.)
    :param id: string  The id parameter specifies a comma-separated list of
        comment thread IDs for the resources that should be retrieved.
    :param videoId: string  The videoId parameter instructs the API to return
        comment threads associated with the specified video ID.

    *Optional parameters:*

    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param moderationStatus: string  This parameter can only be used in a
        properly authorized request. Set this parameter to limit the returned
        comment threads to a particular moderation state.Note: This parameter
        is not supported for use in conjunction with the id parameter. The
        default value is published.Acceptable values are: heldForReview –
        Retrieve comment threads that are awaiting review by a moderator. A
        comment thread can be included in the response if the top-level comment
        or at least one of the replies to that comment are awaiting review.
        likelySpam – Retrieve comment threads classified as likely to be spam.
        A comment thread can be included in the response if the top-level
        comment or at least one of the replies to that comment is considered
        likely to be spam. published – Retrieve threads of published comments.
        This is the default value. A comment thread can be included in the
        response if its top-level comment has been published.
    :param order: string  The order parameter specifies the order in which the
        API response should list comment threads. Valid values are: time -
        Comment threads are ordered by time. This is the default
        behavior.relevance - Comment threads are ordered by relevance.Note:
        This parameter is not supported for use in conjunction with the id
        parameter.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken property identifies the next page of the result that can
        be retrieved.Note: This parameter is not supported for use in
        conjunction with the id parameter.
    :param searchTerms: string  The searchTerms parameter instructs the API to
        limit the API response to only contain comments that contain the
        specified search terms.Note: This parameter is not supported for use in
        conjunction with the id parameter.
    :param textFormat: string  Set this parameter's value to html or plainText
        to instruct the API to return the comments left by users in html
        formatted or in plain text. The default value is html.Acceptable values
        are: html – Returns the comments in HTML format. This is the default
        value. plainText – Returns the comments in plain text format.
    """
    args = locals()
    part_params = {'id', 'replies', 'snippet'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [allThreadsRelatedToChannelId, channelId, id, videoId]]) != 1:
        raise ValueError("make sure you specify exactly one of ['allThreadsRelatedToChannelId', 'channelId', 'id', 'videoId']")

    base_url = 'https://www.googleapis.com/youtube/v3/commentThreads'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=100)


def comments_list(key, part, id=None, parentId=None, maxResults=None,
                  pageToken=None, textFormat=None):
    """Returns a list of comments that match the API request parameters.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more comment resource properties that the API response will
        include.The following list contains the part names that you can include
        in the parameter value and the quota cost for each part: id: 0 snippet:
        1

    *Filters (specify exactly one of the following parameters):*

    :param id: string  The id parameter specifies a comma-separated list of
        comment IDs for the resources that are being retrieved. In a comment
        resource, the id property specifies the comment's ID.
    :param parentId: string  The parentId parameter specifies the ID of the
        comment for which replies should be retrieved. Note: YouTube currently
        supports replies only for top-level comments. However, replies to
        replies may be supported in the future.

    *Optional parameters:*

    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken property identifies the next page of the result that can
        be retrieved.Note: This parameter is not supported for use in
        conjunction with the id parameter.
    :param textFormat: string  This parameter indicates whether the API should
        return comments formatted as HTML or as plain text. The default value
        is html.Acceptable values are: html – Returns the comments in HTML
        format. This is the default value. plainText – Returns the comments in
        plain text format.
    """
    args = locals()
    part_params = {'id', 'snippet'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [id, parentId]]) != 1:
        raise ValueError("make sure you specify exactly one of ['id', 'parentId']")

    base_url = 'https://www.googleapis.com/youtube/v3/comments'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=100)


def channel_sections_list(key, part, channelId=None, id=None, mine=None,
                          hl=None, onBehalfOfContentOwner=None):
    """Returns a list of   resources that match the API request criteria.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more channelSection resource properties that the API response
        will include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a channelSection resource, the snippet property contains
        other properties, such as a display title for the section. If you set
        part=snippet, the API response will also contain all of those nested
        properties.The following list contains the part names that you can
        include in the parameter value and the quota cost for each part:
        contentDetails: 2 id: 0 localizations: 2 snippet: 2 targeting: 2

    *Filters (specify exactly one of the following parameters):*

    :param channelId: string  The channelId parameter specifies a YouTube
        channel ID. If a request specifies a value for this parameter, the API
        will only return the specified channel's sections.
    :param id: string  The id parameter specifies a comma-separated list of IDs
        that uniquely identify the channelSection resources that are being
        retrieved. In a channelSection resource, the id property specifies the
        section's ID.
    :param mine: boolean  This parameter can only be used in a properly
        authorized request. Set this parameter's value to true to retrieve a
        feed of the channel sections associated with the authenticated user's
        YouTube channel.

    *Optional parameters:*

    :param hl: string  The hl parameter instructs the API to retrieve localized
        resource metadata for a specific application language that the YouTube
        website supports. The parameter value must be a language code included
        in the list returned by the i18nLanguages.list method.If localized
        resource details are available in that language, the resource's
        snippet.localized object will contain the localized values. However, if
        localized details are not available, the snippet.localized object will
        contain resource details in the resource's default language.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    """
    args = locals()
    part_params = {'contentDetails', 'id', 'snippet', 'localizations', 'targeting'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [channelId, id, mine]]) != 1:
        raise ValueError("make sure you specify exactly one of ['channelId', 'id', 'mine']")

    base_url = 'https://www.googleapis.com/youtube/v3/channelSections'
    return _combine_requests(args, base_url, count=None, max_allowed=None)


def channels_list(key, part, categoryId=None, forUsername=None, id=None,
                  managedByMe=None, mine=None, mySubscribers=None, hl=None,
                  maxResults=None, onBehalfOfContentOwner=None,
                  pageToken=None):
    """Returns a collection of zero or more   resources that match the request
    criteria.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies a comma-separated list of
        one or more channel resource properties that the API response will
        include.If the parameter identifies a property that contains child
        properties, the child properties will be included in the response. For
        example, in a channel resource, the contentDetails property contains
        other properties, such as the uploads properties. As such, if you set
        part=contentDetails, the API response will also contain all of those
        nested properties.The following list contains the part names that you
        can include in the parameter value and the quota cost for each part:
        auditDetails: 4 brandingSettings: 2 contentDetails: 2
        contentOwnerDetails: 2 id: 0 invideoPromotion: 2 (deprecated)
        localizations: 2 snippet: 2 statistics: 2 status: 2 topicDetails: 2

    *Filters (specify exactly one of the following parameters):*

    :param categoryId: string  The categoryId parameter specifies a YouTube
        guide category, thereby requesting YouTube channels associated with
        that category.
    :param forUsername: string  The forUsername parameter specifies a YouTube
        username, thereby requesting the channel associated with that username.
    :param id: string  The id parameter specifies a comma-separated list of the
        YouTube channel ID(s) for the resource(s) that are being retrieved. In
        a channel resource, the id property specifies the channel's YouTube
        channel ID.
    :param managedByMe: boolean  This parameter can only be used in a properly
        authorized request. Note: This parameter is intended exclusively for
        YouTube content partners.Set this parameter's value to true to instruct
        the API to only return channels managed by the content owner that the
        onBehalfOfContentOwner parameter specifies. The user must be
        authenticated as a CMS account linked to the specified content owner
        and onBehalfOfContentOwner must be provided.
    :param mine: boolean  This parameter can only be used in a properly
        authorized request. Set this parameter's value to true to instruct the
        API to only return channels owned by the authenticated user.
    :param mySubscribers: boolean  This parameter has been deprecated. This
        parameter can only be used in a properly authorized request. Use the
        subscriptions.list method and its mySubscribers parameter to retrieve a
        list of subscribers to the authenticated user's channel.

    *Optional parameters:*

    :param hl: string  The hl parameter instructs the API to retrieve localized
        resource metadata for a specific application language that the YouTube
        website supports. The parameter value must be a language code included
        in the list returned by the i18nLanguages.list method.If localized
        resource details are available in that language, the resource's
        snippet.localized object will contain the localized values. However, if
        localized details are not available, the snippet.localized object will
        contain resource details in the resource's default language.
    :param maxResults: unsigned integer  The maxResults parameter specifies the
        maximum number of items that should be returned in the result set.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The CMS account
        that the user authenticates with must be linked to the specified
        YouTube content owner.
    :param pageToken: string  The pageToken parameter identifies a specific
        page in the result set that should be returned. In an API response, the
        nextPageToken and prevPageToken properties identify other pages that
        could be retrieved.
    """
    args = locals()
    part_params = {'contentDetails', 'id', '(deprecated) localizations', 'snippet', 'auditDetails', 'statistics', 'status', 'invideoPromotion', 'brandingSettings', 'contentOwnerDetails', 'topicDetails'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))
    if sum([bool(p) for p in [categoryId, forUsername, id, managedByMe, mine, mySubscribers]]) != 1:
        raise ValueError("make sure you specify exactly one of ['categoryId', 'forUsername', 'id', 'managedByMe', 'mine', 'mySubscribers']")

    base_url = 'https://www.googleapis.com/youtube/v3/channels'
    return _combine_requests(args, base_url, count=maxResults, max_allowed=50)


def captions_list(key, part, videoId, id=None, onBehalfOfContentOwner=None):
    """Returns a list of caption tracks that are associated with a specified
    video. Note that the API response does not contain the actual captions and
    that the  captions.download  method provides the ability to retrieve a
    caption track.

    *Required parameters:*

    :param key: string  Your Google API key.
    :param part: string  The part parameter specifies the caption resource
        parts that the API response will include.The list below contains the
        part names that you can include in the parameter value and the quota
        cost for each part: id: 0 snippet: 1
    :param videoId: string  The videoId parameter specifies the YouTube video
        ID of the video for which the API should return caption tracks.

    *Optional parameters:*

    :param id: string  The id parameter specifies a comma-separated list of IDs
        that identify the caption resources that should be retrieved. Each ID
        must identify a caption track associated with the specified video.
    :param onBehalfOfContentOwner: string  This parameter can only be used in a
        properly authorized request. Note: This parameter is intended
        exclusively for YouTube content partners.The onBehalfOfContentOwner
        parameter indicates that the request's authorization credentials
        identify a YouTube CMS user who is acting on behalf of the content
        owner specified in the parameter value. This parameter is intended for
        YouTube content partners that own and manage many different YouTube
        channels. It allows content owners to authenticate once and get access
        to all their video and channel data, without having to provide
        authentication credentials for each individual channel. The actual CMS
        account that the user authenticates with must be linked to the
        specified YouTube content owner.
    """
    args = locals()
    part_params = {'id', 'snippet'}
    if not set(part.split(",")).issubset(part_params):
        raise ValueError("make sure your `part` parameter is one or more of " + str(part_params))

    base_url = 'https://www.googleapis.com/youtube/v3/captions'
    return _combine_requests(args, base_url, count=None, max_allowed=None)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import os; os.listdir()
with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
#    'Click>=6.0',
    'pandas',
    'pyasn1',
    'scrapy',
    'twython',
]

setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest',
]

setup(
    author="Elias Dabbas",
    author_email='eliasdabbas@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Productivity and analysis tools for online marketing",
#    entry_points={
#        'console_scripts': [
#            'advertest=advertest.cli:main',
#        ],
#    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='advertising marketing search-engine-optimization adwords '
             'seo sem bingads keyword-research',
    name='advertools',
    packages=find_packages(include=['advertools']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/eliasdabbas/advertools',
    version='0.10.5',
    zip_safe=False,
)



"""
.. _twitter:

Twitter Data API
================
"""

import logging
from functools import wraps

from twython import Twython
import pandas as pd
if int(pd.__version__[0]) >= 1:
    from pandas import json_normalize
else:
    from pandas.io.json import json_normalize


TWITTER_LOG_FMT = ('%(asctime)s | %(levelname)s | %(filename)s:%(lineno)d '
                   '| %(funcName)s | %(message)s')
logging.basicConfig(format=TWITTER_LOG_FMT)

# Functions that depend on 'previous_cursor' and 'next_cursor' to
# navigate requests with a lot of data, request pagination basically.
CURSORED_FUNCTIONS = [
    'get_followers_ids',
    'get_followers_list',
    'get_friends_ids',
    'get_friends_list',
    'get_list_members',
    'get_list_memberships',
    'get_list_subscribers',
    'get_list_subscriptions',
    'get_retweeters_ids',
    'show_owned_lists',
]


# Responses that contain a special key (and the name of that key)
# containing the required data and need to be extracted through
# that key, as opposed to other responses where you can easily
# call DataFrame on them directly
SPECIAL_KEY_FUNCS = {
    'search': 'statuses',
    'get_followers_list': 'users',
    'get_friends_list': 'users',
    'get_list_members': 'users',
    'get_list_subscribers': 'users',
    'get_list_memberships': 'lists',
    'get_list_subscriptions': 'lists',
    'show_owned_lists': 'lists',

}


# Functions that contain an embedded ``user`` key, containing
# 40+ attributes of the user tweeting, listed, retweeted, etc.
USER_DATA_EMBEDDED = {
    'get_favorites': 'tweet_',
    'get_home_timeline': 'tweet_',
    'get_list_memberships': 'list_',
    'get_list_statuses': 'tweet_',
    'get_list_subscriptions': '',
    'get_mentions_timeline': 'tweet_',
    'get_retweets': 'tweet_',
    'get_user_timeline': 'tweet_',
    'lookup_status': 'tweet_',
    'retweeted_of_me': 'tweet_',
    'search': 'tweet_',
    'show_lists': 'list_',
    'show_owned_lists': 'list_',
}


DEFAULT_COUNTS = {
    'get_favorites': 200,
    'get_followers_ids': 5000,
    'get_followers_list': 200,
    'get_friends_ids': 5000,
    'get_friends_list': 200,
    'get_home_timeline': 200,
    'get_list_members': 5000,
    'get_list_memberships': 1000,
    'get_list_statuses': 100,
    'get_list_subscribers': 5000,
    'get_list_subscriptions': 1000,
    'get_mentions_timeline': 200,
    'get_retweeters_ids': 100,
    'get_retweets': 100,
    'get_user_timeline': 200,
    'lookup_status': 100,
    'lookup_user': 100,
    'retweeted_of_me': 100,
    'search': 100,
    'search_users': 20,
    'show_lists': 100,
    'show_owned_lists': 1000
}


def _expand_entities(df):
    if 'tweet_entities' in df:
        colnames = ['tweet_entities_' + x for x in ['mentions', 'hashtags',
                                                    'urls', 'symbols',
                                                    'media']]
        entities_df = json_normalize(df['tweet_entities'])
        mentions = [', '.join(['@' + x['screen_name'] for x in y])
                    for y in entities_df['user_mentions']]
        hashtags = [', '.join(['#' + x['text'] for x in y])
                    for y in entities_df['hashtags']]
        urls = [', '.join([x['expanded_url'] for x in y])
                for y in entities_df['urls']]
        symbols = [', '.join(['$' + x['text'] for x in y])
                   for y in entities_df['symbols']]

        if 'media' in entities_df:
            entities_df['media'] = entities_df['media'].fillna('')
            media = [', '.join([x['media_url'] for x in y]) if y != '' else
                     y for y in entities_df['media']]
            entity_cols = [mentions, hashtags, urls, symbols, media]
        else:
            entity_cols = [mentions, hashtags, urls, symbols]
        col_idx = df.columns.get_loc('tweet_entities')
        for j, col in enumerate(entity_cols):
            df.insert(col_idx+j+1, colnames[j], col)
    return df


def _get_counts(number=None, default=None):
    """Split a number into a list of divisors and the remainder.
    The divisor is the default count in this case."""
    if not number:
        number = 1
    div = divmod(number, default)
    result = [default for x in range(div[0])]
    if div[1] != 0:
        return result + [div[1]]
    return result


def make_dataframe(func):
    @wraps(func)
    def wrapper(count=None, max_id=None, *args, **kwargs):
        nonlocal func

        twtr = Twython(**wrapper.get_auth_params())
        fname = func.__name__
        func = eval('twtr.' + fname)

        if count is None:
            count = DEFAULT_COUNTS[fname]
        counts = _get_counts(count, DEFAULT_COUNTS[fname])

        responses = []
        for i, count in enumerate(counts):
            if fname == 'search':
                if responses and not responses[-1]['statuses']:
                    break
                max_id = (max_id or None) if i == 0 else (responses[-1]['statuses'][-1]['id'] - 1)
            if (fname != 'search') and (fname not in CURSORED_FUNCTIONS):
                if responses and len(responses[-1]) == 0:
                    break
                max_id = (max_id or None) if i == 0 else (responses[-1][-1]['id'] - 1)
            if fname in CURSORED_FUNCTIONS:
                cursor = None if i == 0 else responses[-1]['next_cursor']
                max_id = None
            else:
                cursor = None
            kwargs_log = ', '.join([k + '=' + str(v) for k, v in kwargs.items()])
            args_log = ', '.join(args)
            logging.info(msg=fname + ' | ' + 'Requesting: ' +
                         'count=' + str(count) + ', max_id=' +
                         str(max_id) + ', ' + kwargs_log + args_log)

            resp = func(count=count,
                        max_id=max_id,
                        cursor=cursor,
                        *args, **kwargs)
            responses.append(resp)

        if '_ids' in fname:
            finallist = []
            for sublist in responses:
                finallist.extend(sublist['ids'])
            finaldict = {'previous_cursor': responses[0]['previous_cursor'],
                         'next_cursor': responses[-1]['next_cursor'],
                         'ids': finallist}
            return finaldict

        final_df = pd.DataFrame()
        for resp in responses:
            if SPECIAL_KEY_FUNCS.get(fname):
                resp_df = pd.DataFrame(resp[SPECIAL_KEY_FUNCS.get(fname)])
                if fname in USER_DATA_EMBEDDED:
                    resp_df.columns = [USER_DATA_EMBEDDED[fname] + col for col in resp_df.columns]
                    user_df = pd.DataFrame([x['user'] for x in resp[SPECIAL_KEY_FUNCS.get(fname)]])
                    user_df.columns = ['user_' + col for col in user_df.columns]
                    temp_df = pd.concat([resp_df, user_df], axis=1, sort=False)
                else:
                    temp_df = resp_df
            else:
                resp_df = pd.DataFrame(resp)

                if fname in USER_DATA_EMBEDDED:
                    resp_df.columns = [USER_DATA_EMBEDDED[fname] + x for x in resp_df.columns]
                    user_df = pd.DataFrame([x['user'] for x in resp])
                    user_df.columns = ['user_' + x for x in user_df.columns]
                    temp_df = pd.concat([resp_df, user_df], axis=1)
                else:
                    temp_df = resp_df
            final_df = final_df.append(temp_df, sort=False, ignore_index=True)

        for col in final_df:
            if 'created_at' in col:
                final_df[col] = pd.to_datetime(final_df[col])
        for col in final_df:
            if 'source' in col:
                final_df[col + '_url'] = final_df[col].str.extract('<a href="(.*)" rel=')[0]
                final_df[col] = final_df[col].str.extract('nofollow">(.*)</a>')[0]
        if 'tweet_entities' in final_df:
            return _expand_entities(final_df)

        return final_df
    return wrapper


def authenticate(func):
    """Used internally, please use set_auth_params for authentication."""
    auth_params = {}
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    def set_auth_params(**params):
        nonlocal auth_params
        auth_params.update(params)

    def get_auth_params():
        return auth_params

    wrapper.set_auth_params = set_auth_params
    wrapper.get_auth_params = get_auth_params
    return wrapper


@authenticate
def get_application_rate_limit_status(consumed_only=True):
    """
    Returns the current rate limits for methods belonging to the
        specified resource families.

    :param consumed_only: Whether or not to return only items that
        have been consumed. Otherwise returns the full list.

    https://developer.twitter.com/en/docs/developer-utilities/rate-limit-status/api-reference/get-application-rate_limit_status
    """
    twtr = Twython(**get_application_rate_limit_status.get_auth_params())
    ratelimit = twtr.get_application_rate_limit_status()
    limit_df = pd.DataFrame()
    for resource in ratelimit['resources']:
        temp_df = pd.DataFrame(ratelimit['resources'][resource]).T
        limit_df = limit_df.append(temp_df, sort=False)
    limit_df['reset'] = pd.to_datetime(limit_df['reset'], unit='s')
    limit_df['resource'] = limit_df.index.str.split('/').str[1]
    limit_df.index.name = 'endpoint'
    limit_df = limit_df.sort_values(['resource'])
    limit_df = limit_df.reset_index()
    if consumed_only:
        print(' '*12, 'Rate limit as of:',
              pd.Timestamp.now(tz='UTC').strftime('%Y-%m-%-d %H:%M:%S'))
        return limit_df[limit_df['limit'].ne(limit_df['remaining'])]
    return limit_df


@authenticate
def get_available_trends():
    """
    Returns the locations that Twitter has trending topic information for.

    https://developer.twitter.com/en/docs/trends/locations-with-trending-topics/api-reference/get-trends-available
    """
    twtr = Twython(**get_available_trends.get_auth_params())

    available_trends = twtr.get_available_trends()
    trends_df = pd.DataFrame(available_trends)
    trends_df['code'] = [x['code'] for x in trends_df['placeType']]
    trends_df['place_type'] = [x['name'] for x in trends_df['placeType']]
    del trends_df['placeType']
    trends_df = trends_df.sort_values(['country', 'place_type', 'name'])
    trends_df = trends_df.reset_index(drop=True)
    return trends_df


@make_dataframe
@authenticate
def get_favorites(user_id=None, screen_name=None, count=None, since_id=None,
                  max_id=None, include_entities=None, tweet_mode=None):
    """
    Returns the 20 most recent Tweets favorited by the authenticating
        or specified user.

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param include_entities: (bool - optional) The entities node will be
        omitted when set to False .
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/post-and-engage/api-reference/get-favorites-list
    """
    pass


@make_dataframe
@authenticate
def get_followers_ids(user_id=None, screen_name=None, cursor=None,
                      stringify_ids=None, count=None):
    """
    Returns a cursored collection of user IDs for every user
        following the specified user.

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param cursor: (cursor - semi-optional) Causes the list of connections to
        be broken into pages of no more than 5000 IDs at a time. The number of IDs
        returned is not guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will be
        assumed, which is the first “page.” The response from the API will include a
        previous_cursor and next_cursor to allow paging back and forth. See Using
        cursors to navigate collections for more information.
    :param stringify_ids: (bool - optional) Some programming environments will
        not consume Twitter IDs due to their size. Provide this option to have IDs
        returned as strings instead. More about Twitter IDs.
    :param count: (int - optional) Specifies the number of results to retrieve.

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-followers-ids
    """
    pass


@make_dataframe
@authenticate
def get_followers_list(user_id=None, screen_name=None, cursor=None, count=None,
                       skip_status=None, include_user_entities=None):
    """
    Returns a cursored collection of user objects for users
        following the specified user.

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param cursor: (cursor - semi-optional) Causes the results to be broken
        into pages. If no cursor is provided, a value of -1 will be assumed, which is
        the first “page.” The response from the API will include a previous_cursor and
        next_cursor to allow paging back and forth. See Using cursors to navigate
        collections for more information.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param skip_status: (bool - optional) When set to True, statuses will not
        be included in the returned user objects. If set to any other value, statuses
        will be included.
    :param include_user_entities: (bool - optional) The user object entities
        node will not be included when set to False.

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-followers-list
    """
    pass


@make_dataframe
@authenticate
def get_friends_ids(user_id=None, screen_name=None, cursor=None,
                    stringify_ids=None, count=None):
    """
    Returns a cursored collection of user IDs for every user the
        specified user is following (otherwise known as their "friends").

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param cursor: (cursor - semi-optional) Causes the list of connections to
        be broken into pages of no more than 5000 IDs at a time. The number of IDs
        returned is not guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will be
        assumed, which is the first “page.” The response from the API will include a
        previous_cursor and next_cursor to allow paging back and forth. See Using
        cursors to navigate collections for more information.
    :param stringify_ids: (bool - optional) Some programming environments will
        not consume Twitter IDs due to their size. Provide this option to have IDs
        returned as strings instead. More about Twitter IDs.
    :param count: (int - optional) Specifies the number of results to retrieve.

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-friends-ids
    """
    pass


@make_dataframe
@authenticate
def get_friends_list(user_id=None, screen_name=None, cursor=None, count=None,
                     skip_status=None, include_user_entities=None):
    """
    Returns a cursored collection of user objects for every user the
        specified user is following (otherwise known as their "friends").

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param cursor: (cursor - semi-optional) Causes the results to be broken
        into pages. If no cursor is provided, a value of -1 will be assumed, which is
        the first “page.” The response from the API will include a previous_cursor and
        next_cursor to allow paging back and forth. See Using cursors to navigate
        collections for more information.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param skip_status: (bool - optional) When set to True statuses will not be
        included in the returned user objects.
    :param include_user_entities: (bool - optional) The user object entities
        node will not be included when set to False.

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-friends-list
    """
    pass


@make_dataframe
@authenticate
def get_home_timeline(count=None, since_id=None, max_id=None, trim_user=None,
                      exclude_replies=None, include_entities=None, tweet_mode=None):
    """
    Returns a collection of the most recent Tweets and retweets
        posted by the authenticating user and the users they follow.

    :param count: (int - optional) Specifies the number of results to retrieve.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param trim_user: (bool - optional) When set to True, each Tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param exclude_replies: (bool - optional) This parameter will prevent
        replies from appearing in the returned timeline. Using exclude_replies with the
        count parameter will mean you will receive up-to count Tweets — this is because
        the count parameter retrieves that many Tweets before filtering out retweets
        and replies.
    :param include_entities: (bool - optional) The entities node will not be
        included when set to False.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-home_timeline
    """
    pass


@make_dataframe
@authenticate
def get_list_members(list_id=None, slug=None, owner_screen_name=None, owner_id=None,
                     count=None, cursor=None, include_entities=None, skip_status=None):
    """
    Returns the members of the specified list.

    :param list_id: (str - required) The numerical id of the list.
    :param slug: (str - required) You can identify a list by its slug instead
        of its numerical id. If you decide to do so, note that you’ll also have to
        specify the list owner using the owner_id or owner_screen_name parameters.
    :param owner_screen_name: (str - optional) The screen name of the user who
        owns the list being requested by a slug.
    :param owner_id: (int - optional) The user ID of the user who owns the list
        being requested by a slug.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - semi-optional) Causes the collection of list
        members to be broken into “pages” of consistent sizes (specified by the count
        parameter). If no cursor is provided, a value of -1 will be assumed, which is
        the first “page.” The response from the API will include a previous_cursor and
        next_cursor to allow paging back and forth. See Using cursors to navigate
        collections for more information.
    :param include_entities: (bool - optional) The entities node will not be
        included when set to False.
    :param skip_status: (bool - optional) When set to True statuses will not be
        included in the returned user objects.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-members
    """
    pass


@make_dataframe
@authenticate
def get_list_memberships(user_id=None, screen_name=None, count=None,
                         cursor=None, filter_to_owned_lists=None):
    """
    Returns the lists the specified user has been added to.

    :param user_id: (int - optional) The ID of the user for whom to return
        results. Helpful for disambiguating when a valid user ID is also a valid screen
        name.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results. Helpful for disambiguating when a valid screen name is also
        a user ID.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - optional) Breaks the results into pages. Provide a
        value of -1 to begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in the list.
        It is recommended to always use cursors when the method supports them. See
        Cursoring for more information.
    :param filter_to_owned_lists: (bool - optional) When True, will return just
        lists the authenticating user owns, and the user represented by user_id or
        screen_name is a member of.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-memberships
    """
    pass


@make_dataframe
@authenticate
def get_list_statuses(list_id=None, slug=None, owner_screen_name=None, owner_id=None,
                      since_id=None, max_id=None, count=None, include_entities=None,
                      include_rts=None, tweet_mode=None):
    """
    Returns a timeline of tweets authored by members of the specified list.

    :param list_id: (str - required) The numerical id of the list.
    :param slug: (str - required) You can identify a list by its slug instead
        of its numerical id. If you decide to do so, note that you’ll also have to
        specify the list owner using the owner_id or owner_screen_name parameters.
    :param owner_screen_name: (str - optional) The screen name of the user who
        owns the list being requested by a slug .
    :param owner_id: (int - optional) The user ID of the user who owns the list
        being requested by a slug .
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param include_entities: (bool - optional) Entities are ON by default in
        API 1.1, each tweet includes a node called “entities”. This node offers a
        variety of metadata about the tweet in a discreet structure, including:
        user_mentions, urls, and hashtags. You can omit entities from the result by
        using include_entities=False
    :param include_rts: (bool - optional) When set to True, the list timeline
        will contain native retweets (if they exist) in addition to the standard stream
        of tweets. The output format of retweeted tweets is identical to the
        representation you see in home_timeline.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-statuses
    """
    pass


@make_dataframe
@authenticate
def get_list_subscribers(list_id=None, slug=None, owner_screen_name=None, owner_id=None,
                         count=None, cursor=None, include_entities=None, skip_status=None):
    """
    Returns the subscribers of the specified list.

    :param list_id: (str - required) The numerical id of the list.
    :param slug: (str - required) You can identify a list by its slug instead
        of its numerical id. If you decide to do so, note that you’ll also have to
        specify the list owner using the owner_id or owner_screen_name parameters.
    :param owner_screen_name: (str - optional) The screen name of the user who
        owns the list being requested by a slug .
    :param owner_id: (int - optional) The user ID of the user who owns the list
        being requested by a slug .
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - optional) Breaks the results into pages. A single
        page contains 20 lists. Provide a value of -1 to begin paging. Provide values
        as returned in the response body’s next_cursor and previous_cursor attributes
        to page back and forth in the list. See Using cursors to navigate collections
        for more information.
    :param include_entities: (bool - optional) When set to True, each tweet
        will include a node called “entities”. This node offers a variety of metadata
        about the tweet in a discreet structure, including: user_mentions, urls, and
        hashtags. While entities are opt-in on timelines at present, they will be made
        a default component of output in the future. See Tweet Entities for more
        details.
    :param skip_status: (bool - optional) When set to Truestatuses will not be
        included in the returned user objects.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-subscribers
    """
    pass


@make_dataframe
@authenticate
def get_list_subscriptions(user_id=None, screen_name=None, count=None,
                           cursor=None):
    """
    Obtain a collection of the lists the specified user is subscribed to.

    :param user_id: (int - optional) The ID of the user for whom to return
        results. Helpful for disambiguating when a valid user ID is also a valid screen
        name.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results. Helpful for disambiguating when a valid screen name is also
        a user ID.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - optional) Breaks the results into pages. Provide a
        value of -1 to begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in the list.
        It is recommended to always use cursors when the method supports them. See
        Cursoring for more information.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-subscriptions
    """
    pass


@make_dataframe
@authenticate
def get_mentions_timeline(count=None, since_id=None, max_id=None,
                          trim_user=None, include_entities=None, tweet_mode=None):
    """
    Returns the 20 most recent mentions (tweets containing a users's
        @screen_name) for the authenticating user.

    :param count: (int - optional) Specifies the number of results to retrieve.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param trim_user: (bool - optional) When set to True, each tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param include_entities: (bool - optional) The entities node will not be
        included when set to False.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-mentions_timeline
    """
    pass


@authenticate
def get_place_trends(ids, exclude=None):
    """
    Returns the top 10 trending topics for a specific WOEID, if
        trending information is available for it.

    :param id: (int or list of ints - required) run ``get_available_trends()`` for 
        the full listing.
        The Yahoo! Where On Earth ID of the 
        location to return trending information for. Global information is available 
        by using 1 as the WOEID .
    :param exclude: (str - optional) Setting this equal to hashtags will remove
        all hashtags from the trends list.

    https://developer.twitter.com/en/docs/trends/trends-for-location/api-reference/get-trends-place
    """
    twtr = Twython(**get_place_trends.get_auth_params())
    trends_df = pd.DataFrame()
    if isinstance(ids, int):
        ids = [ids]
    for place_id in ids:

        place_trends = twtr.get_place_trends(id=place_id)
        trend_df = pd.DataFrame(place_trends[0]['trends'])
        trend_df = trend_df.sort_values(['tweet_volume'], ascending=False)
        trend_df['location'] = place_trends[0]['locations'][0]['name']
        trend_df['woeid'] = place_trends[0]['locations'][0]['woeid']
        trend_df['time'] = pd.to_datetime(place_trends[0]['created_at'])

        trends_df = trends_df.append(trend_df, ignore_index=True)

    trends_df = trends_df.sort_values(['woeid', 'tweet_volume'],
                                      ascending=[True, False])
    trends_df = trends_df.reset_index(drop=True)
    available = get_available_trends()
    available = available[['country', 'parentid', 'woeid', 'place_type']]
    final_df = pd.merge(trends_df, available, on='woeid')
    final_df['local_rank'] = (final_df
                              .groupby('woeid')['tweet_volume']
                              .rank(method='dense', ascending=False))
    final_df['country_rank'] = (final_df
                                .groupby('country')['tweet_volume']
                                .rank(method='dense', ascending=False))
    final_df = final_df[['name', 'location', 'tweet_volume', 'local_rank',
                         'country', 'country_rank', 'time','place_type',
                         'promoted_content', 'woeid', 'parentid']]
    return final_df



@make_dataframe
@authenticate
def get_retweeters_ids(id, count=None, cursor=None, stringify_ids=None):
    """
    Returns a collection of up to 100 user IDs belonging to users who
        have retweeted the tweet specified by the ``id`` parameter.
        It's better to use get_retweets because passing a count > 100
        will only get you duplicated data. 100 is the maximum even
        if there were more retweeters.

    :param id: (int - required) The numerical ID of the desired status.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - semi-optional) Causes the list of IDs to be broken
        into pages of no more than 100 IDs at a time. The number of IDs returned is not
        guaranteed to be 100 as suspended users are filtered out after connections are
        queried. If no cursor is provided, a value of -1 will be assumed, which is the
        first “page.” The response from the API will include a previous_cursor and
        next_cursor to allow paging back and forth. See our cursor docs for more
        information. While this method supports the cursor parameter, the entire result
        set can be returned in a single cursored collection. Using the count parameter
        with this method will not provide segmented cursors for use with this
        parameter.
    :param stringify_ids: (bool - optional) Many programming environments will
        not consume Tweet ids due to their size. Provide this option to have ids
        returned as strings instead.

    https://developer.twitter.com/en/docs/tweets/post-and-engage/api-reference/get-statuses-retweeters-ids
    """
    pass


@make_dataframe
@authenticate
def get_retweets(id, trim_user=None, tweet_mode=None):
    """
    Returns up to 100 of the first retweets of a given tweet.

    :param id: (int - required) The numerical ID of the desired status.
    :param trim_user: (bool - optional) When set to True, each tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/post-and-engage/api-reference/post-statuses-retweet-id
    """
    pass



@authenticate
def get_supported_languages():
    """
    Returns the list of languages supported by Twitter along with
        their ISO 639-1 code.

    https://developer.twitter.com/en/docs/developer-utilities/supported-languages/api-reference/get-help-languages
    """
    twtr = Twython(**get_supported_languages.get_auth_params())
    langs = twtr.get_supported_languages()
    return pd.DataFrame(langs)



@make_dataframe
@authenticate
def get_user_timeline(user_id=None, screen_name=None, since_id=None,
                      count=None, max_id=None, trim_user=None, exclude_replies=None,
                      include_rts=None, tweet_mode=None):
    """
    Returns a collection of the most recent Tweets posted by the user
        indicated by the ``screen_name`` or ``user_id`` parameters.

    :param user_id: (int - optional) The ID of the user for whom to return
        results.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets that can be accessed through the API. If the limit of Tweets has occured
        since the since_id, the since_id will be forced to the oldest ID available.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param trim_user: (bool - optional) When set to True, each Tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param exclude_replies: (bool - optional) This parameter will prevent
        replies from appearing in the returned timeline. Using exclude_replies with the
        count parameter will mean you will receive up-to count tweets — this is because
        the count parameter retrieves that many Tweets before filtering out retweets
        and replies.
    :param include_rts: (bool - optional) When set to False , the timeline will
        strip any native retweets (though they will still count toward both the maximal
        length of the timeline and the slice selected by the count parameter). Note: If
        you’re using the trim_user parameter in conjunction with include_rts, the
        retweets will still contain a full user object.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/timelines/api-reference/get-statuses-user_timeline
    """
    pass


@make_dataframe
@authenticate
def lookup_status(id, include_entities=None, trim_user=None, map=None,
                  include_ext_alt_text=None, include_card_uri=None, tweet_mode=None):
    """
    Returns fully-hydrated tweet objects for up to 100 tweets per
        request, as specified by comma-separated values passed to the ``id``
        parameter.

    :param id: (int - required) A comma separated list of Tweet IDs, up to 100
        are allowed in a single request.
    :param include_entities: (bool - optional) The entities node that may
        appear within embedded statuses will not be included when set to False.
    :param trim_user: (bool - optional) When set to True, each Tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param map: (bool - optional) When using the map parameter, Tweets that do
        not exist or cannot be viewed by the current user will still have their key
        represented but with an explicitly null value paired with it
    :param include_ext_alt_text: (bool - optional) If alt text has been added
        to any attached media entities, this parameter will return an ext_alt_text
        value in the top-level key for the media entity. If no value has been set, this
        will be returned as null
    :param include_card_uri: (bool - optional) When set to True, each Tweet
        returned will include a card_uri attribute when there is an ads card attached
        to the Tweet and when that card was attached using the card_uri value.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/post-and-engage/api-reference/get-statuses-lookup
    """
    pass


@make_dataframe
@authenticate
def lookup_user(screen_name=None, user_id=None, include_entities=None,
                tweet_mode=None):
    """
    Returns fully-hydrated user objects for up to 100 users per request,
        as specified by comma-separated values passed to the ``user_id`` and/or
        ``screen_name`` parameters.

    :param screen_name: (str - optional) A comma separated list of screen
        names, up to 100 are allowed in a single request. You are strongly encouraged
        to use a POST for larger (up to 100 screen names) requests.
    :param user_id: (int - optional) A comma separated list of user IDs, up to
        100 are allowed in a single request. You are strongly encouraged to use a POST
        for larger requests.
    :param include_entities: (bool - optional) The entities node that may
        appear within embedded statuses will not be included when set to False.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-users-lookup
    """
    pass


@make_dataframe
@authenticate
def retweeted_of_me(count=None, since_id=None, max_id=None, trim_user=None,
                    include_entities=None, include_user_entities=None, tweet_mode=None):
    """
    Returns the most recent tweets authored by the authenticating user
        that have been retweeted by others.

    :param count: (int - optional) Specifies the number of results to retrieve.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param trim_user: (bool - optional) When set to True, each tweet returned
        in a timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    :param include_entities: (bool - optional) The tweet entities node will not
        be included when set to False .
    :param include_user_entities: (bool - optional) The user entities node will
        not be included when set to False .
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    https://developer.twitter.com/en/docs/tweets/post-and-engage/api-reference/get-statuses-retweets_of_me
    """
    pass


@make_dataframe
@authenticate
def search(q, geocode=None, lang=None, locale=None, result_type=None,
           count=None, until=None, since_id=None, max_id=None, include_entities=None,
           tweet_mode=None):
    """
    Returns a collection of relevant Tweets matching a specified query.

    :param q: (str - required) A UTF-8, URL-encoded search query of 500
        characters maximum, including operators. Queries may additionally be limited by
        complexity.
    :param geocode: (lat lon dist - optional) Returns tweets by users located
        within a given radius of the given latitude/longitude. The location is
        preferentially taking from the Geotagging API, but will fall back to their
        Twitter profile. The parameter value is specified by ”
        latitude,longitude,radius “, where radius units must be specified as either ”
        mi ” (miles) or ” km ” (kilometers). Note that you cannot use the near operator
        via the API to geocode arbitrary locations; however you can use this geocode
        parameter to search near geocodes directly. A maximum of 1,000 distinct “sub-
        regions” will be considered when using the radius modifier.
    :param lang: (str - optional) Restricts tweets to the given language, given
        by an ISO 639-1 code. Language detection is best-effort.
    :param locale: (str - optional) Specify the language of the query you are
        sending (only ja is currently effective). This is intended for language-
        specific consumers and the default should work in the majority of cases.
    :param result_type: (str - optional) Optional. Specifies what type of
        search results you would prefer to receive. The current default is “mixed.”
        Valid values include: * mixed : Include both popular and real time results in
        the response. * recent : return only the most recent results in the response *
        popular : return only the most popular results in the response.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param until: (date - optional) Returns tweets created before the given
        date. Date should be formatted as YYYY-MM-DD. Keep in mind that the search
        index has a 7-day limit. In other words, no tweets will be found for a date
        older than one week.
    :param since_id: (int - optional) Returns results with an ID greater than
        (that is, more recent than) the specified ID. There are limits to the number of
        Tweets which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest ID
        available.
    :param max_id: (int - optional) Returns results with an ID less than (that
        is, older than) or equal to the specified ID.
    :param include_entities: (bool - optional) The entities node will not be
        included when set to False.
    :param tweet_mode: (str - optional) Valid request values are compat and
        extended, which give compatibility mode and extended mode, respectively for
        Tweets that contain over 140 characters

    +------------------------------------+-----------------------------------------------------+
    |                           Operator | Finds Tweets...                                     |
    +====================================+=====================================================+
    |                      watching now  | containing both “watching” and “now”. This is the   |
    |                                    | default operator.                                   |
    +------------------------------------+-----------------------------------------------------+
    |                      “happy hour”  |  containing the exact phrase “happy hour”.          |
    +------------------------------------+-----------------------------------------------------+
    |                      love OR hate  |  containing either “love” or “hate” (or both).      |
    +------------------------------------+-----------------------------------------------------+
    |                        beer -root  |  containing “beer” but not “root”.                  |
    +------------------------------------+-----------------------------------------------------+
    |                            #haiku  |  containing the hashtag “haiku”.                    |
    +------------------------------------+-----------------------------------------------------+
    |                     from:interior  |  sent from Twitter account “interior”.              |
    +------------------------------------+-----------------------------------------------------+
    | list:NASA/astronauts-in-space-now  | sent from a Twitter account in the NASA list        |
    |                                    | astronauts-in-space-now                             |
    +------------------------------------+-----------------------------------------------------+
    |                           to:NASA  | a Tweet authored in reply to Twitter account        |
    |                                    | “NASA”.                                             |
    +------------------------------------+-----------------------------------------------------+
    |                             @NASA  |  mentioning Twitter account “NASA”.                 |
    +------------------------------------+-----------------------------------------------------+
    |              politics filter:safe  | containing “politics” with Tweets marked as         |
    |                                    | potentially sensitive removed.                      |
    +------------------------------------+-----------------------------------------------------+
    |                puppy filter:media  |  containing “puppy” and an image or video.          |
    +------------------------------------+-----------------------------------------------------+
    |            puppy -filter:retweets  |  containing “puppy”, filtering out retweets         |
    +------------------------------------+-----------------------------------------------------+
    |         puppy filter:native_video  | containing “puppy” and an uploaded video, Amplify   |
    |                                    | video, Periscope, or Vine.                          |
    +------------------------------------+-----------------------------------------------------+
    |            puppy filter:periscope  |  containing “puppy” and a Periscope video URL.      |
    +------------------------------------+-----------------------------------------------------+
    |                 puppy filter:vine  |  containing “puppy” and a Vine.                     |
    +------------------------------------+-----------------------------------------------------+
    |               puppy filter:images  | containing “puppy” and links identified as photos,  |
    |                                    | including third parties such as Instagram.          |
    +------------------------------------+-----------------------------------------------------+
    |                puppy filter:twimg  | containing “puppy” and a pic.twitter.com link       |
    |                                    | representing one or more photos.                    |
    +------------------------------------+-----------------------------------------------------+
    |            hilarious filter:links  |  containing “hilarious” and linking to URL.         |
    +------------------------------------+-----------------------------------------------------+
    |                  puppy url:amazon  | containing “puppy” and a URL with the word          |
    |                                    | “amazon” anywhere within it.                        |
    +------------------------------------+-----------------------------------------------------+
    |        superhero since:2015-12-21  | containing “superhero” and sent since date          |
    |                                    | “2015-12-21” (year-month-day).                      |
    +------------------------------------+-----------------------------------------------------+
    |            puppy until:2015-12-21  | containing “puppy” and sent before the date         |
    |                                    | “2015-12-21”.                                       |
    +------------------------------------+-----------------------------------------------------+
    |                   movie -scary :)  | containing “movie”, but not “scary”, and with a     |
    |                                    | positive attitude.                                  |
    +------------------------------------+-----------------------------------------------------+
    |                         flight :(  | containing “flight” and with a negative attitude.   |
    +------------------------------------+-----------------------------------------------------+
    |                          traffic ? | containing “traffic” and asking a question.         |
    +------------------------------------+-----------------------------------------------------+

    https://developer.twitter.com/en/docs/tweets/search/api-reference/get-search-tweets
    """
    pass


@make_dataframe
@authenticate
def search_users(q, page=None, count=None, include_entities=None):
    """
    Provides a simple, relevance-based search interface to public user
        accounts on Twitter.

    :param q: (str - required) The search query to run against people search.
    :param page: (int - optional) Specifies the page of results to retrieve.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param include_entities: (bool - optional) The entities node will not be
        included in embedded Tweet objects when set to False .

    https://developer.twitter.com/en/docs/accounts-and-users/follow-search-get-users/api-reference/get-users-search
    """
    pass


@make_dataframe
@authenticate
def show_lists(user_id=None, screen_name=None, reverse=None):
    """
    Returns all lists the authenticating or specified user subscribes to,
        including their own.

    :param user_id: (int - optional) The ID of the user for whom to return
        results. Helpful for disambiguating when a valid user ID is also a valid screen
        name. Note: : Specifies the ID of the user to get lists from. Helpful for
        disambiguating when a valid user ID is also a valid screen name.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results. Helpful for disambiguating when a valid screen name is also
        a user ID.
    :param reverse: (bool - optional) Set this to true if you would like owned
        lists to be returned first. See description above for information on how this
        parameter works.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-list
    """
    pass


@make_dataframe
@authenticate
def show_owned_lists(user_id=None, screen_name=None, count=None, cursor=None):
    """
    Returns the lists owned by the specified Twitter user.

    :param user_id: (int - optional) The ID of the user for whom to return
        results. Helpful for disambiguating when a valid user ID is also a valid screen
        name.
    :param screen_name: (str - optional) The screen name of the user for whom
        to return results. Helpful for disambiguating when a valid screen name is also
        a user ID.
    :param count: (int - optional) Specifies the number of results to retrieve.
    :param cursor: (cursor - optional) Breaks the results into pages. Provide a
        value of -1 to begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in the list.
        It is recommended to always use cursors when the method supports them. See
        Cursoring for more information.

    https://developer.twitter.com/en/docs/accounts-and-users/create-manage-lists/api-reference/get-lists-ownerships
    """
    pass


FUNCTIONS = [
    get_application_rate_limit_status,
    get_available_trends,
    get_favorites,
    get_followers_ids,
    get_followers_list,
    get_friends_ids,
    get_friends_list,
    get_home_timeline,
    get_list_members,
    get_list_memberships,
    get_list_statuses,
    get_list_subscribers,
    get_list_subscriptions,
    get_mentions_timeline,
    get_place_trends,
    get_retweeters_ids,
    get_retweets,
    get_supported_languages,
    get_user_timeline,
    lookup_status,
    lookup_user,
    retweeted_of_me,
    search,
    search_users,
    show_lists,
    show_owned_lists,
]


def set_auth_params(app_key=None, app_secret=None, oauth_token=None,
                    oauth_token_secret=None, access_token=None,
                    token_type='bearer', oauth_version=1, api_version='1.1',
                    client_args=None, auth_endpoint='authenticate'):
    """The main function for authentication. 
    Needs to be called once in a session. 
    
    First you need to create a developer account and app:
    https://developer.twitter.com/ to get your credentials. 
    
    Different ways to authenticate:
    https://twython.readthedocs.io/en/latest/usage/starting_out.html
    """
    params = locals()
    for func in FUNCTIONS:
        func.set_auth_params(**params)
    return None


logging.getLogger().setLevel('INFO')
